/**********************************************************************************************
 * Program Name:  Acme Filters
 *
 * Module Name:   Acme Filters AE.c
 *
 * Description:	  Apple Event code for Acme Filters.
 *
 * Author:        Eric V. Curtis
 *
 * History:
 *        v1.0    29-May-1995  Created.
 *
 * Functions:     InstallAEHandlers()
 *                GotRequiredParams()
 *                DoRequiredSuite()
 *                HandleOpenDoc()
 **********************************************************************************************/



/************ I N C L U D E   F I L E S *******************************************************/

#include <stdio.h>

#include <AppleEvents.h>
#include <GestaltEqu.h>

#include "Acme Filters AE.h"
#include "Acme Filters.h"
#include "ASCToMac.h"
#include "Delete RF.h"
#include "FindOriginal.h"
#include "Rot 13.h"
#include "Tab2Space.h"
#include "Touch.h"
#include "TypeCreator.h"



/**********************************************************************************************
 * Function Name:  InstallAEHandlers
 *
 * Description:    Install our apple event handlers.
 *
 * Parameters:     None.
 *
 * Return Value:   noErr if ok, error code otherwise.
 **********************************************************************************************/
void	InstallAEHandlers( void )
{
	// required events
	AEInstallEventHandler( kCoreEventClass, kAEOpenApplication, NewAEEventHandlerProc( DoRequiredSuite ), 1, false );
	AEInstallEventHandler( kCoreEventClass, kAEOpenDocuments,  NewAEEventHandlerProc( DoRequiredSuite ), 2, false );
	AEInstallEventHandler( kCoreEventClass, kAEQuitApplication,  NewAEEventHandlerProc( DoRequiredSuite ), 4, false );
}



/**********************************************************************************************
 * Function Name:  GotRequiredParams
 *
 * Description:    Check for missed, but required, apple event parameters.
 *
 * Parameters:     event - the apple event currently being processed.
 *
 * Return Value:   noErr - no parameters were missed.
 *                 errAEParamMissed - at least one parameter was missed.
 *                 err - some other error occurred.
 **********************************************************************************************/
OSErr	GotRequiredParams( AppleEvent *event )
{
	OSErr		err;
	DescType	returnedType;
	Size		actualSize;
	
	err = AEGetAttributePtr( event, keyMissedKeywordAttr, typeWildCard,
							 &returnedType, (Ptr)0L, (Size)0L, &actualSize );
	
	if ( err == errAEDescNotFound )
	{
		err = noErr;					// we got all the data
	}
	else if ( err == noErr )
	{
		err = errAEParamMissed;			// we missed some data
	}
	
	return ( err );
}



/**********************************************************************************************
 * Function Name:  DoRequiredSuite
 *
 * Description:    Handle the open and quit application apple events.
 *
 * Parameters:     Standard stuff.
 *
 * Return Value:   noErr if ok, error code otherwise.
 **********************************************************************************************/
pascal OSErr	DoRequiredSuite( AppleEvent *event, AppleEvent *reply, long refCon )
{
#pragma unused ( reply )
	if ( refCon == 1 )					// kAEOpenApplication
	{
		AboutBox( kStartTimeOut );
		gAutoQuit = false;				// never autoquit if app was double-clicked
	}
	
	if ( refCon == 2 )					// kAEOpenDocuments
	{
		HandleOpenDoc( event );
		if ( gAutoQuit ) gQuit = true;	// only autoquit if app launched via dragged files
	}
	
	if ( refCon == 4 )					// kAEQuitApplication
	{
		gQuit = true;
	}
	
	return ( noErr );
}



/**********************************************************************************************
 * Function Name:  HandleOpenDoc
 *
 * Description:    Process an 'Open Documents' apple event.
 *
 * Parameters:     event  - the apple event to be processed
 *
 * Return Value:   noErr - everything went OK
 *                 other - something strange is afoot
 **********************************************************************************************/
pascal OSErr	HandleOpenDoc( AppleEvent *event )
{
	OSErr		err;
	AEDescList	docList;
	long		count;
	long		i;
	long		ticks;
	OSType		type;
	OSType		selType, selCrea;
	unsigned long	created = 0, modified = 0;
	
	AEKeyword	keyword;
	DescType	returnedType;
	FSSpec		file;
	Size		grow, actualSize;
	
	// compact our heap
	MaxMem( &grow );
	
	err = AEGetParamDesc( event, keyDirectObject, typeAEList, &docList );
	if ( err ) goto exit;
	
	err = GotRequiredParams( event );
	if ( err ) goto fail;
	
	err = AECountItems( &docList, &count );
	if ( err ) goto fail;
	
	// let the user choose a filter if the option key is down
	if ( SpecialKeyDown( 0x3a ) && !gFromMenu )
	{
		err = ChooseDefault( false );
	}
	if ( err ) goto fail;
	
	// get the first file in the list
	err = AEGetNthPtr( &docList, 1, typeFSS, &keyword, &returnedType, &file, sizeof( FSSpec ), &actualSize );
	if ( err ) goto fail;
	
	// let the user specify a date if necessary
	if ( (gConversionType == kTouch) && SpecialKeyDown( 0x3a ) )
	{
		err = PickDateTime( &file, &created, &modified );
	}
	if ( err ) goto fail;
	
	// let the user specify a type and creator if necessary
	if ( gConversionType == kTypeCreator )
	{
		err = PickTypeCreator( &file, &selType, &selCrea );
	}
	if ( err ) goto fail;
	
	// always clear the cancel flag
	gCancelReq = false;
	
	// process each file in the list
	if ( gConversionType != kFindOriginal) Progress( eCreate, count, nil, 0 );
	Delay( 12, &ticks );
	for ( i = 1; i <= count; i++ )
	{
		err = AEGetNthPtr( &docList, i, typeFSS, &keyword, &returnedType, &file, sizeof( FSSpec ), &actualSize );
		if ( err ) goto fail;
		
		type = GetFileType( &file );
		
		Progress( eFileName, 0, file.name, 0 );
		switch ( gConversionType )
		{
			case kTypeCreator:
				TypeCreator( &file, selType, selCrea );
				break;
			
			case kTouch:
				Touch( &file, created, modified );
				break;
			
			case kTab2Space:
				if ( type == kTEXTType ) Tab2Space( &file );
				break;
			
			case kRot13:
				if ( type == kTEXTType ) Rot13( &file );
				break;
			
			case kFindOriginal:
				FindOriginal( &file );
				break;
			
			case kDeleteRF:
				if ( type == kTEXTType ) DeleteRF( &file );
				break;
			
			case kASCToMac:
			default:
				if ( type == kTEXTType ) ASCToMac( &file );
				break;
		}
		Progress( eRemainPercent, count-i, nil, (short)((i/(float)count)*100.0) );
		
		if ( gCancelReq ) i = count+1;
	}
	Progress( eFileName, 0, "\p", 0 );
	Delay( 12, &ticks );
	Progress( eDelete, 0, nil, 0 );
	
fail:
	AEDisposeDesc( &docList );
	
	// reset the default conversion type
	gConversionType = LoadPreferred();
	gFromMenu = false;
	
exit:
	return ( err );
}
